@@ -5,13 +5,13 @@
  * Copyright (C) 2010-2012 Canonical Ltd.
  * Copyright (C) 2013 Red Hat, Inc.
  *
- * This library is free software: you can redistribute it and/or modify it 
+ * This library is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License version 3
  * as published by the Free Software Foundation.
  *
- * This library is distributed in the hope that it will be useful, but 
- * WITHOUT ANY WARRANTY; without even the implied warranties of 
- * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
+ * This library is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranties of
+ * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
  * PURPOSE.  See the GNU Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
@@ -54,18 +54,12 @@
 #include <unistd.h>
 
 #include "version.h"
-#include "event-names.h"
 
 /* File format version we write out
    NOTE: if you bump the version number, make sure you update README */
 #define EVEMU_FILE_MAJOR 1
 #define EVEMU_FILE_MINOR 2
 
-#ifndef UI_SET_PROPBIT
-#define UI_SET_PROPBIT		_IOW(UINPUT_IOCTL_BASE, 110, int)
-#define EVIOCGPROP(len)		_IOC(_IOC_READ, 'E', 0x09, len)
-#endif
-
 #define SYSCALL(call) while (((call) == -1) && (errno == EINTR))
 
 enum error_level {
@@ -120,20 +114,17 @@ static int next_line(FILE *fp, char **line, size_t *sz)
 	return 0;
 }
 
-static void copy_bits(unsigned char *mask, const unsigned long *bits, int bytes)
-{
-	int i;
-	for (i = 0; i < bytes; i++) {
-		int pos = 8 * (i % sizeof(long));
-		mask[i] = (bits[i / sizeof(long)] >> pos) & 0xff;
-	}
-}
 
 struct evemu_device *evemu_new(const char *name)
 {
 	struct evemu_device *dev = calloc(1, sizeof(struct evemu_device));
 
 	if (dev) {
+		dev->evdev = libevdev_new();
+		if (!dev->evdev) {
+			free(dev);
+			return NULL;
+		}
 		dev->version = EVEMU_VERSION;
 		evemu_set_name(dev, name);
 	}
@@ -143,6 +134,9 @@ struct evemu_device *evemu_new(const char *name)
 
 void evemu_delete(struct evemu_device *dev)
 {
+	if (dev->uidev)
+		evemu_destroy(dev);
+	libevdev_free(dev->evdev);
 	free(dev);
 }
 
@@ -153,186 +147,146 @@ unsigned int evemu_get_version(const struct evemu_device *dev)
 
 const char *evemu_get_name(const struct evemu_device *dev)
 {
-	return dev->name;
+	return libevdev_get_name(dev->evdev);
 }
 
 void evemu_set_name(struct evemu_device *dev, const char *name)
 {
-	if (name && strlen(name) < sizeof(dev->name))
-		strcpy(dev->name, name);
+	if (name)
+		libevdev_set_name(dev->evdev, name);
 }
 
-unsigned int evemu_get_id_bustype(const struct evemu_device *dev)
-{
-	return dev->id.bustype;
+#define ID_GETTER(field) \
+unsigned int evemu_get_id_##field(const struct evemu_device *dev) { \
+	return libevdev_get_id_##field(dev->evdev); \
 }
 
-void evemu_set_id_bustype(struct evemu_device *dev,
-			  unsigned int bustype)
-{
-	dev->id.bustype = bustype;
+ID_GETTER(bustype)
+ID_GETTER(version)
+ID_GETTER(product)
+ID_GETTER(vendor)
+
+#define ID_SETTER(field) \
+void evemu_set_id_##field(struct evemu_device *dev, unsigned int value) { \
+	libevdev_set_id_##field(dev->evdev, value); \
 }
 
-unsigned int evemu_get_id_vendor(const struct evemu_device *dev)
-{
-	return dev->id.vendor;
+ID_SETTER(bustype)
+ID_SETTER(version)
+ID_SETTER(product)
+ID_SETTER(vendor)
+
+#define ABS_GETTER(field) \
+int evemu_get_abs_##field(const struct evemu_device *dev, int code) { \
+	return libevdev_get_abs_##field(dev->evdev, code); \
 }
 
-void evemu_set_id_vendor(struct evemu_device *dev,
-			 unsigned int vendor)
-{
-	dev->id.vendor = vendor;
+ABS_GETTER(minimum)
+ABS_GETTER(maximum)
+ABS_GETTER(fuzz)
+ABS_GETTER(flat)
+ABS_GETTER(resolution)
+
+#define ABS_SETTER(field) \
+void evemu_set_abs_##field(struct evemu_device *dev, int code, int value) { \
+	libevdev_set_abs_##field(dev->evdev, code, value); \
 }
 
-unsigned int evemu_get_id_product(const struct evemu_device *dev)
-{
-	return dev->id.product;
-}
-
-void evemu_set_id_product(struct evemu_device *dev,
-			  unsigned int product)
-{
-	dev->id.product = product;
-}
-
-unsigned int evemu_get_id_version(const struct evemu_device *dev)
-{
-	return dev->id.version;
-}
-
-void evemu_set_id_version(struct evemu_device *dev,
-			  unsigned int version)
-{
-	dev->id.version = version;
-}
-
-int evemu_get_abs_minimum(const struct evemu_device *dev, int code)
-{
-	return dev->abs[code].minimum;
-}
-
-void evemu_set_abs_minimum(struct evemu_device *dev, int code, int min)
-{
-	dev->abs[code].minimum = min;
-}
-
-int evemu_get_abs_maximum(const struct evemu_device *dev, int code)
-{
-	return dev->abs[code].maximum;
-}
+ABS_SETTER(minimum)
+ABS_SETTER(maximum)
+ABS_SETTER(fuzz)
+ABS_SETTER(flat)
+ABS_SETTER(resolution)
 
 int evemu_get_abs_current_value(const struct evemu_device *dev, int code)
 {
-	return dev->abs[code].value;
-}
-
-void evemu_set_abs_maximum(struct evemu_device *dev, int code, int max)
-{
-	dev->abs[code].maximum = max;
-}
-
-int evemu_get_abs_fuzz(const struct evemu_device *dev, int code)
-{
-	return dev->abs[code].fuzz;
-}
-
-void evemu_set_abs_fuzz(struct evemu_device *dev, int code, int fuzz)
-{
-	dev->abs[code].fuzz = fuzz;
-}
-
-int evemu_get_abs_flat(const struct evemu_device *dev, int code)
-{
-	return dev->abs[code].flat;
-}
-
-void evemu_set_abs_flat(struct evemu_device *dev, int code, int flat)
-{
-	dev->abs[code].flat = flat;
-}
-
-int evemu_get_abs_resolution(const struct evemu_device *dev, int code)
-{
-	return dev->abs[code].resolution;
-}
-
-void evemu_set_abs_resolution(struct evemu_device *dev, int code, int res)
-{
-	dev->abs[code].resolution = res;
+	return libevdev_get_event_value(dev->evdev, EV_ABS, code);
 }
 
 int evemu_has_prop(const struct evemu_device *dev, int code)
 {
-	return (dev->prop[code >> 3] >> (code & 7)) & 1;
+	return libevdev_has_property(dev->evdev, code);
 }
 
 int evemu_has_event(const struct evemu_device *dev, int type, int code)
 {
-	return (dev->mask[type][code >> 3] >> (code & 7)) & 1;
+	return libevdev_has_event_code(dev->evdev, type, code);
 }
 
 int evemu_has_bit(const struct evemu_device *dev, int type)
 {
-	return (dev->mask[0][type >> 3] >> (type & 7)) & 1;
+	return libevdev_has_event_type(dev->evdev, type);
 }
 
 int evemu_extract(struct evemu_device *dev, int fd)
 {
-	unsigned long bits[64];
-	int rc, i;
-
-	memset(dev, 0, sizeof(*dev));
-
-	SYSCALL(rc = ioctl(fd, EVIOCGNAME(sizeof(dev->name)-1), dev->name));
-	if (rc < 0)
-		return rc;
-
-	SYSCALL(rc = ioctl(fd, EVIOCGID, &dev->id));
-	if (rc < 0)
-		return rc;
-
-	SYSCALL(rc = ioctl(fd, EVIOCGPROP(sizeof(bits)), bits));
-	if (rc >= 0) {
-		copy_bits(dev->prop, bits, rc);
-		dev->pbytes = rc;
+	if (libevdev_get_fd(dev->evdev) != -1) {
+		libevdev_free(dev->evdev);
+		dev->evdev = libevdev_new();
+		if (!dev->evdev)
+			return -ENOMEM;
 	}
-
-	for (i = 0; i < EV_CNT; i++) {
-		SYSCALL(rc = ioctl(fd, EVIOCGBIT(i, sizeof(bits)), bits));
-		if (rc < 0)
-			continue;
-		copy_bits(dev->mask[i], bits, rc);
-		dev->mbytes[i] = rc;
-	}
-
-	for (i = 0; i < ABS_CNT; i++) {
-		if (!evemu_has_event(dev, EV_ABS, i))
-			continue;
-		SYSCALL(rc = ioctl(fd, EVIOCGABS(i), &dev->abs[i]));
-		if (rc < 0)
-			return rc;
-	}
-
-	return 0;
+	return libevdev_set_fd(dev->evdev, fd);
 }
 
-static void write_prop(FILE * fp, const unsigned char *mask, int bytes)
+static inline int bit_is_set(unsigned char *mask, int bit)
 {
+	return !!(mask[bit/8] & (1 << (bit & 0x7)));
+}
+
+static inline void set_bit(unsigned char *mask, int bit)
+{
+	mask[bit/8] |= 1 << (bit & 0x7);
+}
+
+#define max(a, b) (a > b) ? a : b
+
+static void write_prop(FILE * fp, const struct evemu_device *dev)
+{
+#ifdef INPUT_PROP_MAX
 	int i;
-	for (i = 0; i < bytes; i += 8)
+	unsigned char mask[max(8, (INPUT_PROP_MAX + 7)/8)] = {0};
+
+	for (i = 0; i < INPUT_PROP_MAX; i ++) {
+		if (evemu_has_prop(dev, i))
+			set_bit(mask, i);
+	}
+
+	for (i = 0; i < (INPUT_PROP_MAX + 7)/8; i +=8) {
 		fprintf(fp, "P: %02x %02x %02x %02x %02x %02x %02x %02x\n",
 			mask[i], mask[i + 1], mask[i + 2], mask[i + 3],
 			mask[i + 4], mask[i + 5], mask[i + 6], mask[i + 7]);
+	}
+#endif
 }
 
-static void write_mask(FILE * fp, int index,
-		       const unsigned char *mask, int bytes)
+static void write_mask(FILE * fp, const struct evemu_device *dev)
 {
-	int i;
-	for (i = 0; i < bytes; i += 8)
-		fprintf(fp, "B: %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
-			index, mask[i], mask[i + 1], mask[i + 2], mask[i + 3],
-			mask[i + 4], mask[i + 5], mask[i + 6], mask[i + 7]);
+	unsigned int type;
+
+	/* Write EV_SYN/SYN_REPORT only. Just because older versions
+	   printed out exactly that */
+	fprintf(fp, "B: 00 0b 00 00 00 00 00 00 00\n");
+
+	for (type = 1 /* don't write EV_SYN */; type < EV_CNT; type++) {
+		int i;
+		int max = libevdev_event_type_get_max(type);
+		unsigned char mask[KEY_CNT] = {0};
+		unsigned int code;
+
+		if (max == -1)
+			continue;
+
+		for (code = 0; code < (unsigned int)max; code++)
+			if (evemu_has_event(dev, type, code))
+				set_bit(mask, code);
+
+		for (i = 0; i < (max + 7)/8; i += 8) {
+			fprintf(fp, "B: %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
+				type, mask[i], mask[i + 1], mask[i + 2], mask[i + 3],
+				mask[i + 4], mask[i + 5], mask[i + 6], mask[i + 7]);
+		}
+	}
 }
 
 static void write_abs(FILE *fp, int index, const struct input_absinfo *abs)
@@ -354,13 +308,13 @@ static void write_desc(const struct evemu_device *dev, FILE *fp)
 		if (!evemu_has_bit(dev, i))
 			continue;
 
-		fprintf(fp, "#   Event type %d (%s)\n", i, event_get_type_name(i));
-		for (j = 0; j < KEY_MAX; j++) {
+		fprintf(fp, "#   Event type %d (%s)\n", i, libevdev_event_type_get_name(i));
+		for (j = 0; j < libevdev_event_type_get_max(i); j++) {
 			if (!evemu_has_event(dev, i, j))
 				continue;
 
 			fprintf(fp, "#     Event code %d (%s)\n",
-				    j, event_get_code_name(i, j));
+				    j, libevdev_event_code_get_name(i, j));
 			if (i == EV_ABS) {
 				fprintf(fp, "#       Value %6d\n"
 					    "#       Min   %6d\n"
@@ -383,7 +337,8 @@ static void write_desc(const struct evemu_device *dev, FILE *fp)
 	for (i = 0; i < INPUT_PROP_MAX; i++) {
 		if (!evemu_has_prop(dev, i))
 			continue;
-		fprintf(fp, "#   Property  type %d (%s)\n", i, input_prop_map[i]);
+		fprintf(fp, "#   Property  type %d (%s)\n", i,
+				libevdev_property_get_name(i));
 	}
 #endif
 }
@@ -404,19 +359,17 @@ int evemu_write(const struct evemu_device *dev, FILE *fp)
 		evemu_get_id_product(dev),
 		evemu_get_id_version(dev));
 
-	write_prop(fp, dev->prop, dev->pbytes);
-
-	for (i = 0; i < EV_CNT; i++)
-		write_mask(fp, i, dev->mask[i], dev->mbytes[i]);
+	write_prop(fp, dev);
+	write_mask(fp, dev);
 
 	for (i = 0; i < ABS_CNT; i++)
 		if (evemu_has_event(dev, EV_ABS, i))
-			write_abs(fp, i, &dev->abs[i]);
+			write_abs(fp, i, libevdev_get_abs_info(dev->evdev, i));
 
 	return 0;
 }
 
-static int parse_name(struct evemu_device *dev, const char *line, struct version *fversion)
+static int parse_name(struct evemu_device *dev, const char *line)
 {
 	int matched;
 	char *devname = NULL;
@@ -435,16 +388,16 @@ static int parse_name(struct evemu_device *dev, const char *line, struct version
 	return matched > 0;
 }
 
-static int parse_bus_vid_pid_ver(struct evemu_device *dev, const char *line, struct version *fversion)
+static int parse_bus_vid_pid_ver(struct evemu_device *dev, const char *line)
 {
 	int matched;
 	unsigned bustype, vendor, product, version;
 
 	if ((matched = sscanf(line, "I: %04x %04x %04x %04x\n",
 				    &bustype, &vendor, &product, &version)) > 0) {
-		evemu_set_id_bustype(dev, bustype),
-		evemu_set_id_vendor(dev, vendor),
-		evemu_set_id_product(dev, product),
+		evemu_set_id_bustype(dev, bustype);
+		evemu_set_id_vendor(dev, vendor);
+		evemu_set_id_product(dev, product);
 		evemu_set_id_version(dev, version);
 	}
 
@@ -454,16 +407,16 @@ static int parse_bus_vid_pid_ver(struct evemu_device *dev, const char *line, str
 	return matched == 4;
 }
 
-static int parse_prop(struct evemu_device *dev, const char *line, struct version *fversion)
+static int parse_prop(struct evemu_device *dev, const char *line)
 {
 	int matched;
-	unsigned int mask[8];
-	int i;
+	unsigned char mask[8];
+	size_t i;
 
 	if (strlen(line) <= 2 || strncmp(line, "P:", 2) != 0)
 		return 0;
 
-	matched = sscanf(line, "P: %02x %02x %02x %02x %02x %02x %02x %02x\n",
+	matched = sscanf(line, "P: %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx\n",
 				mask + 0, mask + 1, mask + 2, mask + 3,
 				mask + 4, mask + 5, mask + 6, mask + 7);
 
@@ -472,22 +425,26 @@ static int parse_prop(struct evemu_device *dev, const char *line, struct version
 		return -1;
 	}
 
-	for (i = 0; i < 8; i++)
-		dev->prop[dev->pbytes++] = mask[i];
+	for (i = 0; i < sizeof(mask) * 8; i++) {
+		if (bit_is_set(mask, i))
+			libevdev_enable_property(dev->evdev, dev->pbytes * 8 + i);
+	}
+
+	dev->pbytes += 8;
 
 	return 1;
 }
 
-static int parse_mask(struct evemu_device *dev, const char *line, struct version *fversion)
+static int parse_mask(struct evemu_device *dev, const char *line)
 {
 	int matched;
-	unsigned int mask[8];
+	unsigned char mask[8];
 	unsigned int index, i;
 
 	if (strlen(line) <= 2 || strncmp(line, "B:", 2) != 0)
 		return 0;
 
-	matched = sscanf(line, "B: %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
+	matched = sscanf(line, "B: %02x %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx\n",
 				&index, mask + 0, mask + 1, mask + 2, mask + 3,
 				mask + 4, mask + 5, mask + 6, mask + 7);
 
@@ -496,8 +453,19 @@ static int parse_mask(struct evemu_device *dev, const char *line, struct version
 		return -1;
 	}
 
-	for (i = 0; i < 8; i++)
-		dev->mask[index][dev->mbytes[index]++] = mask[i];
+	for (i = 0; i < sizeof(mask) * 8; i++) {
+		if (bit_is_set(mask, i)) {
+			struct input_absinfo abs = {0}; /* dummy */
+
+			abs.minimum = 0;
+			abs.maximum = 1;
+
+			unsigned int code = dev->mbytes[index] * 8 + i;
+			libevdev_enable_event_code(dev->evdev, index, code, (index == EV_ABS) ? &abs : NULL);
+		}
+	}
+
+	dev->mbytes[index] += 8;
 
 	return 1;
 }
@@ -560,16 +528,10 @@ int evemu_read(struct evemu_device *dev, FILE *fp)
 	struct version file_version; /* file format version */
 	size_t size = 0;
 	char *line = NULL;
-	char *name = NULL;
 
-	if (strlen(evemu_get_name(dev)) > 0)
-		name = strdup(evemu_get_name(dev));
-	memset(dev, 0, sizeof(*dev));
+	memset(dev->mbytes, 0, sizeof(*dev->mbytes));
+	dev->pbytes = 0;
 	dev->version = EVEMU_VERSION;
-	if (name) {
-		evemu_set_name(dev, name);
-		free(name);
-	}
 
 	/* first line _may_ be version */
 	if (!first_line(fp, &line, &size)) {
@@ -584,13 +546,13 @@ int evemu_read(struct evemu_device *dev, FILE *fp)
 		goto out;
 	}
 
-	if (!parse_name(dev, line, &file_version))
+	if (!parse_name(dev, line))
 		goto out;
 
 	if (!next_line(fp, &line, &size))
 		goto out;
 
-	if (!parse_bus_vid_pid_ver(dev, line, &file_version))
+	if (!parse_bus_vid_pid_ver(dev, line))
 		goto out;
 
 	/* devices without prop/mask/abs bits are valid */
@@ -599,13 +561,13 @@ int evemu_read(struct evemu_device *dev, FILE *fp)
 		goto out;
 	}
 
-	while((rc = parse_prop(dev, line, &file_version)) > 0)
+	while((rc = parse_prop(dev, line)) > 0)
 		if (!next_line(fp, &line, &size))
 			break;
 	if (rc == -1)
 		goto out;
 
-	while((rc = parse_mask(dev, line, &file_version)) > 0)
+	while((rc = parse_mask(dev, line)) > 0)
 		if (!next_line(fp, &line, &size))
 			break;
 	if (rc == -1)
@@ -624,30 +586,36 @@ out:
 	return rc;
 }
 
-static void write_event_desc(FILE *fp, const struct input_event *ev)
+static int write_event_desc(FILE *fp, const struct input_event *ev)
 {
+	int rc;
+
 	if (ev->type == EV_SYN) {
 		if (ev->code == SYN_MT_REPORT)
-			fprintf(fp, "# ++++++++++++ %s (%d) ++++++++++\n",
-				event_get_code_name(ev->type, ev->code),
-				ev->value);
+			rc = fprintf(fp, "# ++++++++++++ %s (%d) ++++++++++\n",
+				     libevdev_event_code_get_name(ev->type, ev->code),
+				     ev->value);
 		else
-			fprintf(fp, "# ------------ %s (%d) ----------\n",
-				event_get_code_name(ev->type, ev->code),
-				ev->value);
+			rc = fprintf(fp, "# ------------ %s (%d) ----------\n",
+				     libevdev_event_code_get_name(ev->type, ev->code),
+				     ev->value);
 	} else {
-		fprintf(fp, "# %s / %-20s %d\n",
-			event_get_type_name(ev->type),
-			event_get_code_name(ev->type, ev->code),
-			ev->value);
+		rc = fprintf(fp, "# %s / %-20s %d\n",
+			     libevdev_event_type_get_name(ev->type),
+			     libevdev_event_code_get_name(ev->type, ev->code),
+			     ev->value);
 	}
+	return rc;
 }
 
 int evemu_write_event(FILE *fp, const struct input_event *ev)
 {
-	return fprintf(fp, "E: %lu.%06u %04x %04x %04d	",
-		       ev->time.tv_sec, (unsigned)ev->time.tv_usec,
-		       ev->type, ev->code, ev->value);
+	int rc;
+	rc = fprintf(fp, "E: %lu.%06u %04x %04x %04d	",
+		     ev->time.tv_sec, (unsigned)ev->time.tv_usec,
+		     ev->type, ev->code, ev->value);
+	rc += write_event_desc(fp, ev);
+	return rc;
 }
 
 static inline long time_to_long(const struct timeval *tv) {
@@ -681,7 +649,6 @@ int evemu_record(FILE *fp, int fd, int ms)
 			time = time_to_long(&ev.time);
 			ev.time = long_to_time(time - offset);
 			evemu_write_event(fp, &ev);
-			write_event_desc(fp, &ev);
 			fflush(fp);
 		}
 	}
@@ -762,7 +729,7 @@ int evemu_play_one(int fd, const struct input_event *ev)
 {
 	int ret;
 	SYSCALL(ret = write(fd, ev, sizeof(*ev)));
-	return (ret < sizeof(*ev));
+	return (ret == -1 || (size_t)ret < sizeof(*ev)) ? -1 : 0;
 }
 
 static void evemu_warn_about_incompatible_event(struct input_event *ev)
@@ -775,8 +742,8 @@ static void evemu_warn_about_incompatible_event(struct input_event *ev)
 			error(WARNING, "You are trying to play events incompatbile with this device. "
 					"Is this the right device/recordings file?\n");
 		error(WARNING, "%s %s is not supported by this device.\n",
-				event_get_type_name(ev->type),
-				event_get_code_name(ev->type, ev->code));
+				libevdev_event_type_get_name(ev->type),
+				libevdev_event_code_get_name(ev->type, ev->code));
 	} else if (warned == max_warnings + 1) {
 		error(INFO, "warned about incompatible events %d times. Will be quiet now.\n",
 				warned - 1);
@@ -812,128 +779,26 @@ int evemu_play(FILE *fp, int fd)
 	return 0;
 }
 
-static int set_prop_bit(int fd, int code)
+int evemu_create(struct evemu_device *dev, int fd)
 {
-	int ret;
-	SYSCALL(ret = ioctl(fd, UI_SET_PROPBIT, code));
-	return ret;
+	return libevdev_uinput_create_from_device(dev->evdev, fd, &dev->uidev);
 }
 
-static int set_event_bit(int fd, int type, int code)
+int evemu_create_managed(struct evemu_device *dev)
 {
-	int ret = 0;
+	return libevdev_uinput_create_from_device(dev->evdev,
+		LIBEVDEV_UINPUT_OPEN_MANAGED, &dev->uidev);
+}
 
-	switch(type) {
-	case EV_SYN:
-		SYSCALL(ret = ioctl(fd, UI_SET_EVBIT, code));
-		break;
-	case EV_KEY:
-		SYSCALL(ret = ioctl(fd, UI_SET_KEYBIT, code));
-		break;
-	case EV_REL:
-		SYSCALL(ret = ioctl(fd, UI_SET_RELBIT, code));
-		break;
-	case EV_ABS:
-		SYSCALL(ret = ioctl(fd, UI_SET_ABSBIT, code));
-		break;
-	case EV_MSC:
-		SYSCALL(ret = ioctl(fd, UI_SET_MSCBIT, code));
-		break;
-	case EV_LED:
-		SYSCALL(ret = ioctl(fd, UI_SET_LEDBIT, code));
-		break;
-	case EV_SND:
-		SYSCALL(ret = ioctl(fd, UI_SET_SNDBIT, code));
-		break;
-	case EV_FF:
-		SYSCALL(ret = ioctl(fd, UI_SET_FFBIT, code));
-		break;
-	case EV_SW:
-		SYSCALL(ret = ioctl(fd, UI_SET_SWBIT, code));
-		break;
+const char *evemu_get_devnode(struct evemu_device *dev)
+{
+	return libevdev_uinput_get_devnode(dev->uidev);
+}
+
+void evemu_destroy(struct evemu_device *dev)
+{
+	if (dev->uidev) {
+		libevdev_uinput_destroy(dev->uidev);
+		dev->uidev = NULL;
 	}
-
-	return ret;
-}
-
-static int set_prop(const struct evemu_device *dev, int fd)
-{
-	int bits = 8 * dev->pbytes;
-	int ret, i;
-	int success = 0;
-	for (i = 0; i < bits; i++) {
-		if (!evemu_has_prop(dev, i))
-			continue;
-		ret = set_prop_bit(fd, i);
-		/* Older kernels aways return errors on UI_SET_PROPBIT.
-		   Assume that if we only get failures, it may be an older
-		   kernel and report success anyway. */
-		if (ret < 0) {
-			if (success)
-				return ret;
-		} else if (!success)
-			success = 1;
-	}
-	return 0;
-}
-
-static int set_mask(const struct evemu_device *dev, int type, int fd)
-{
-	int bits = 8 * dev->mbytes[type];
-	int ret, i;
-	for (i = 0; i < bits; i++) {
-		if (!evemu_has_event(dev, type, i))
-			continue;
-
-		/* kernel doesn't like those */
-		if (type == EV_ABS &&
-			evemu_get_abs_maximum(dev, i) == 0 && evemu_get_abs_minimum(dev, i) == 0)
-			continue;
-
-		ret = set_event_bit(fd, type, i);
-		if (ret < 0)
-			return ret;
-	}
-	return 0;
-}
-
-int evemu_create(const struct evemu_device *dev, int fd)
-{
-	struct uinput_user_dev udev;
-	int ret, i;
-
-	memset(&udev, 0, sizeof(udev));
-	memcpy(udev.name, dev->name, sizeof(udev.name));
-	udev.id = dev->id;
-	for (i = 0; i < ABS_CNT; i++) {
-		if (!evemu_has_event(dev, EV_ABS, i))
-			continue;
-		udev.absmax[i] = evemu_get_abs_maximum(dev, i);
-		udev.absmin[i] = evemu_get_abs_minimum(dev, i);
-		udev.absfuzz[i] = evemu_get_abs_fuzz(dev, i);
-		udev.absflat[i] = evemu_get_abs_flat(dev, i);
-	}
-
-	ret = set_prop(dev, fd);
-	if (ret < 0)
-		return ret;
-
-	for (i = 0; i < EV_CNT; i++) {
-		ret = set_mask(dev, i, fd);
-		if (ret < 0)
-			return ret;
-	}
-
-	SYSCALL(ret = write(fd, &udev, sizeof(udev)));
-	if (ret < 0)
-		return ret;
-
-	SYSCALL(ret = ioctl(fd, UI_DEV_CREATE, NULL));
-	return ret;
-}
-
-void evemu_destroy(int fd)
-{
-	int ret;
-	SYSCALL(ret = ioctl(fd, UI_DEV_DESTROY, NULL));
 }
